#!/usr/bin/env python

# Copyright (c) 2008 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import re
import types
import pprint
import logging

# Komodo modules.
import process

log = logging.getLogger('bzrlib')
#log.setLevel(logging.DEBUG)

#---- exceptions

class BzrLibError(Exception):
    pass

class Bzr:
    """A proxy to the Bzr command line app."""
    def __init__(self, executable='bzr', **options):
        self.executable = executable
        self.optd = options
        # Abortable process handle.
        self._processHelper = process.AbortableProcessHelper()

    def _runCommand(self, argv, cwd=None, env=None, input=None,
                    allowed_exit_codes=None, terminalHandler=None):
        """Prepare and run the given arg vector, 'argv', and return the
        results.  Returns (<stdout>, <stderr>).
        """
        if type(argv) not in (types.ListType, types.TupleType):
            raise BzrLibError("_runCommand:: argv is not a list: %r" % (argv, ))
        log.debug("Running %r in cwd %r", argv, cwd)
        output = None
        if env is None:
            import koprocessutils
            env = koprocessutils.getUserEnv()

        # Komodo can only handle svn messages in english.
        # http://bugs.activestate.com/show_bug.cgi?id=45677
        env['LC_MESSAGES'] = 'en_US'
        # Set LANGUAGE too, otherwise it may still come back in another lang
        # http://bugs.activestate.com/show_bug.cgi?id=68615
        env['LANGUAGE'] = 'en_US'

        # A light wrapper around process.ProcessOpen.
        p = self._processHelper.ProcessOpen(cmd=argv, cwd=cwd, env=env,
                                            universal_newlines=True)
        try:
            if input:
                input = input.encode("utf-8")
            if terminalHandler:
                terminalHandler.hookIO(p.stdin, p.stdout, p.stderr, " ".join(argv))
                p.wait()
                # Output and errors have gone to the terminal.
                output = ''
                error = ''
            else:
                output, error = p.communicate(input)
            retval = p.returncode
            p.close()
        finally:
            self._processHelper.ProcessDone()
        
        if retval != 0:
            if not allowed_exit_codes or retval not in allowed_exit_codes:
                raise BzrLibError("Error code %r running %r:\n%s"
                                  % (retval, argv, ''.join(error)))
        return output, error

    def abort(self):
        self._processHelper.ProcessAbort()

    def add(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "add"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def branch(self, urls, path=None, command_options="",
               terminalHandler=None, env=None):
        if type(urls) in types.StringTypes:
            urls = [urls]
        argv = [self.executable, 'branch'] + urls
        cwd = None
        if path:
            cwd, path = os.path.split(path)
            argv += [path]
        if command_options:
            # Example: "--rev 103 --clean"
            argv += command_options.split(" ")
        output, error = self._runCommand(argv,
                                         cwd=cwd,
                                         env=env,
                                         terminalHandler=terminalHandler)
        result = {}
        result['stdout'] = output
        result['stderr'] = error
        return result

    def commit(self, files, cwd=None, options=None, env=None, msgpath=None):
        result = {}
        argv = [self.executable, "commit"]
        if options:
            argv += options.split(" ")
        if msgpath is not None:
            argv += [ "--file", msgpath ]
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def diff(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "diff"]
        if options:
            argv += options.split(" ")
        if files:
            argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env,
                                         allowed_exit_codes=(1, 2))
        result['stdout'] = output
        result['stderr'] = error
        return result

    def _logItemFrom(self, data):
        logItem = {}
        lines = data.rstrip().splitlines(0)
        lineno = 0
        mode = "parsing"
        file_action = ""
        message = []
        modified_files = []
        added_files = []
        removed_files = []
        for lineno in range(len(lines)):
            line = lines[lineno]
            split = line.split(":", 1)
            if len(split) == 2:
                item_type = split[0]
                if item_type == "revno":
                    logItem['revision'] = split[1].strip()
                elif item_type == "committer":
                    logItem['author'] = split[1].strip()
                elif item_type == "timestamp":
                    logItem['date'] = split[1].strip()
                elif item_type in ("modified", "added", "removed"):
                    mode = "files"
                    file_action = item_type
                    continue
                elif item_type == "message":
                    mode = "message"
                    continue
                elif item_type[0][0] not in " \t":
                    mode = "parsing"
                    continue
            if mode == "message":
                message.append(line[2:])
            elif mode == "files":
                if file_action == "modified":
                    modified_files.append(line.strip())
                elif file_action == "added":
                    added_files.append(line.strip())
                elif file_action == "removed":
                    removed_files.append(line.strip())

        logItem['message'] = "\n".join(message) + "\n"
        logItem['added_files'] = added_files
        logItem['modified_files'] = modified_files
        logItem['removed_files'] = removed_files
        return logItem

    log_splitter_re = re.compile(r'^%s$' % ('-'*60, ), flags=re.MULTILINE)

    def log(self, path, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "log", "-v"]
        if options:
            argv += options.split(" ")
        argv.append(path)
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        log_items = []
        matchIterator = self.log_splitter_re.finditer(output)
        try:
            match = matchIterator.next()
        except StopIteration:
            match = None
        while match:
            text_start = match.end()
            try:
                matchNext = matchIterator.next()
            except StopIteration:
                matchNext = None
            if matchNext:
                message = output[text_start:matchNext.start()]
            else:
                message = output[text_start:]

            log_items.append(self._logItemFrom(message))
            match = matchNext

        result['log'] = log_items
        return result

    def remove(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "remove"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def revert(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "revert"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    status_re = re.compile(r"^(?P<status_flags>.)(?P<status>.)\s+(?P<path>.*)$")
    def status(self, files=None, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "status", "-S"]
        if options:
            argv += options.split(" ")
        if files:
            argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        files = {}
        for line in output.splitlines(0):
            match = self.status_re.match(line)
            if match:
                hit = match.groupdict()
                path = hit['path']
                fullpath = os.path.normpath(os.path.join(cwd, path))
                files[fullpath] = { 'status': hit['status'],
                                    'status_flags': hit['status_flags'],
                                    'relpath': path }
        result['files'] = files
        return result

    def statusEx(self, cwd=None, options=None, env=None):
        # bzr does not give status information on files that are in the tree but
        # are not unmodified (most files), so we perform a "bzr ls" to get the
        # rest of the files we are missing. The "--versioned" ensures the files
        # that get printed are under version control, the only ones we want.
        result = self.status(files=None, cwd=cwd, options=options, env=env)
        output, error = self._runCommand([self.executable, "ls", "--versioned"],
                                         cwd=cwd, env=env)
        result['stdout'] += output
        result['stderr'] += error

        files = result.get('files')
        for line in output.splitlines(0):
            relpath = line.strip()
            fullpath = os.path.normpath(os.path.join(cwd, relpath))
            if fullpath not in files:
                files[fullpath] = { 'status': 'H',
                                    'status_flags': ' ',
                                    'relpath': relpath,}
        return result

    def update(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "update"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result
