/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: JavaScript; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

// The status bar API for Komodo.
//
// Here is an excerpt from the status bar requirements document:
//  --------------------------------------------------------------------------
//  |  <message>                  | <ck> | <enc.> | <pos.> | <sel.> | <lang.> /,
//  --------------------------------------------------------------------------
//     ^--(a) Message                ^       ^        ^         ^       ^    ^
//        (b) Linter/Checker status -'       |        |         |       |    |
//        (c) Encoding ----------------------'        |         |       |    |
//        (d) Position (line and column) -------------'         |       |    |
//        (e) Selection (lines and chars) ----------------------'       |    |
//        (f) Language -------------------------------------------------'    |
//        (z) Grippy --------------------------------------------------------'
//
// Implementation:
//
// The Status Bar observes various status notification messages (e.g.
// 'current_view_encoding_changed') via Mozilla's nsIObserver mechanism.
// Because some of the
// status bar segments are specific to the currently focused view ((b)-(f)),
// the preference is that notifications only be sent for changes to the current
// view. This is not always easy, so in some cases the logic in this module
// must ensure that notifications are for the current view.
//
// How to make a status message:
// Method 1.
//      If you are in JavaScript code you can just issue a
//      ko.statusBar.AddMessage(...) call. See that function for a description
//      of the arguments.
// Method 2.
//      You can send a 'status_message' notification via the nsIObserver
//      mechanism where the 'subject' argument of the Notify() is a
//      koIStatusMessage object. Here is an example in Python code:
//
//        obsSvc = components.classes["@mozilla.org/observer-service;1"]\
//                 .getService(components.interfaces.nsIObserverService)
//        sm = components.classes["@activestate.com/koStatusMessage;1"]\
//             .createInstance(components.interfaces.koIStatusMessage)
//        sm.category = "<some string>"
//        sm.msg = "<some string>"
//        sm.timeout = 0    # 0 for no timeout, else a number of milliseconds
//        sm.highlight = 0  # boolean, whether or not to highlight
//        try:
//            obsSvc.notifyObservers(sm, "status_message", None)
//        except COMException, ex:
//            pass
//
// TODO:
//  - double-click and right-click functionality as per the requirements doc
//  - read-only status icon
//  - better icons for showing check status (more obvious, the inprogress one
//    must at least be replaced)


//---- globals
if (typeof(ko) == 'undefined') {
    var ko = {};
}
ko.statusBar = {};
(function() { /* ko.statusBar */

var _log = ko.logging.getLoggingMgr().getLogger('statusbar');
var _messageStack = Components.classes["@activestate.com/koStatusMessageStack;1"].
                          createInstance(Components.interfaces.koIStatusMessageStack);
var _observer = null;
var _prefObserver = null;
var _updateLintMessageTimer = null;
var _addMessageTimer = null;

//_log.setLevel(ko.logging.LOG_DEBUG);


//---- helper functions

function _updateEncoding(view) {
    if (typeof(view)=='undefined' || !view || !view.document)
        return;
    try {
        //XXX It would probably be cleaner to add an "encoding_name"
        //    attribute to koIView and then let each view type override that.
        //    Then the view-startpage could return null to indicate N/A.
        if (view.getAttribute("type") == "startpage") {
            _clearEncoding();
        } else {
            var encoding = view.document.encoding.short_encoding_name;
            var encodingWidget = document.getElementById('statusbar-encoding');
            encodingWidget.setAttribute("label", encoding);
        }
    } catch(e) {
        _clearEncoding();
    }
}


function _clearEncoding() {
    var encodingWidget = document.getElementById('statusbar-encoding');
    encodingWidget.removeAttribute("label");
}


function _updateLanguage(view) {
    if (typeof(view)=='undefined' || !view || !view.document)
        return;
    try {
        //XXX It would probably be cleaner to handle the "startpage language
        //    is N/A" logic in the view system, but I don't know how to
        //    easily do that right now.
        var languageWidget = document.getElementById('statusbar-language-menu');
        if (view.getAttribute("type") == "startpage") {
            _clearLanguage();
            languageWidget.setAttribute('collapsed', 'true');
        } else {
            var language = view.document.language;
            languageWidget.setAttribute("label", language);
            languageWidget.removeAttribute('collapsed');
        }
    } catch(e) {
        _clearLanguage();
    }
}


function _clearLanguage() {
    var languageWidget = document.getElementById('statusbar-language-menu');
    languageWidget.setAttribute("label", "");
}

function _updateLintMessage(view) {
    // The timeout has been called, remove the setTimeout id
    _updateLintMessageTimer = null;
    if (!view || !view.document || !view.lintBuffer) {
        return;
    }

    // If there are lint result messages for the current line then display
    // them.
    var lintResults = view.lintBuffer.lintResults;
    var lintError = view.lintBuffer.errorString;
    var lintMessage = null;
    if (!lintError && lintResults) {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResultsAtPosition(view.currentLine,
                                         view.currentColumn,
                                         resultsObj,
                                         numResultsObj);
        var results = resultsObj.value;
        var numResults = numResultsObj.value;

        for (var i = 0; i < numResults; i++) {
            if (lintMessage) {
                lintMessage += "\n" + results[i].description;
            } else {
                lintMessage = results[i].description;
            }
        }
    }

    if (lintMessage) {
        _addMessage(lintMessage, "check", 0, false);
    } else {
        _addMessage(null, "check", 0, false);
    }
}



var _updateSection_cache = null; // [view.document.displayPath, line number]

/**
 * Update the 'Sections' statusbar panel label.
 */
function _updateSectionPanel(view, forceUpdate /* =false */) {
    if (typeof(forceUpdate) == "undefined") forceUpdate = false;
    
    // If this isn't a supported view type, then clear the curr section.
    if (!view || view.getAttribute("type") != "editor") {
        _disableSectionPanel();
        return;
    }

    // Guard against doing more work here than necessary.
    if (!forceUpdate
        && _updateSection_cache
        && view.document.displayPath == _updateSection_cache[0]
        && view.currentLine == _updateSection_cache[1]) {
        return;
    }
    _updateSection_cache = [view.document.displayPath, view.currentLine];

    var section = view.document.ciBuf.curr_section_from_line(view.currentLine);

    if (section == null) {
        _clearSectionPanel();
    } else {
        var section_panel = document.getElementById("statusbar-section");
        var section_image = document.getElementById("statusbar-section-image");
        var section_label = document.getElementById("statusbar-section-label");
        var section_arrows = document.getElementById("statusbar-section-arrows");
        var section_text = document.getElementById("statusbar-section-textbox");
        var strbundle = document.getElementById("komodo_strings");

        section_panel.removeAttribute("disabled");
        section_text.removeAttribute("disabled");
        section_arrows.removeAttribute("collapsed");
        section_label.setAttribute("tooltiptext",
            strbundle.getFormattedString("sectionTooltip",
                [section.title, section.lang, section.type]));
        section_label.setAttribute("value", section.title);

        // Update the 'codeintel-type-$type' class.
        var classes = section_image.getAttribute("class").split(/\s+/);
        var new_classes = [];
        for (var i in classes) {
            if (classes[i].slice(0, 15) != "codeintel-type-") {
                new_classes.push(classes[i]);
            }
        }
        new_classes.push("codeintel-type-"+section.type);
        section_image.setAttribute("class", new_classes.join(' '));
    }
}


/**
 * Clear the 'Sections' statusbar panel label.
 */
function _clearSectionPanel() {
    var section_label = document.getElementById("statusbar-section-label");
    if (section_label.value == "\u2013") {
        return; // Already cleared.
    }

    var section_panel = document.getElementById("statusbar-section");
    var section_image = document.getElementById("statusbar-section-image");
    var section_arrows = document.getElementById("statusbar-section-arrows");

    section_panel.removeAttribute("disabled");
    section_arrows.removeAttribute("collapsed");
    section_label.setAttribute("tooltiptext", "Sections");
    section_image.removeAttribute("src");
    section_label.setAttribute("value", "\u2013");  // &mdash;

    // Remove the 'codeintel-type-$type' class.
    var classes = section_image.getAttribute("class").split(/\s+/);
    var new_classes = [];
    for (var i in classes) {
        if (classes[i].slice(0, 15) != "codeintel-type-") {
            new_classes.push(classes[i]);
        }
    }
    section_image.setAttribute("class", new_classes.join(' '));
}

/**
 * Disable the 'Sections' statusbar panel.
 */
function _disableSectionPanel() {
    _updateSection_cache = null;
    
    var section_panel = document.getElementById("statusbar-section");
    var section_image = document.getElementById("statusbar-section-image");
    var section_label = document.getElementById("statusbar-section-label");
    var section_arrows = document.getElementById("statusbar-section-arrows");
    var section_text = document.getElementById("statusbar-section-textbox");
    
    section_panel.setAttribute("disabled", "true");
    section_text.setAttribute("disabled", "true");
    section_arrows.setAttribute("collapsed", "true");
    
    section_label.setAttribute("tooltiptext", "Sections");
    section_image.removeAttribute("src");
    section_label.setAttribute("value", "");

    // Remove the 'codeintel-type-$type' class.
    var classes = section_image.getAttribute("class").split(/\s+/);
    var new_classes = [];
    for (var i in classes) {
        if (classes[i].slice(0, 15) != "codeintel-type-") {
            new_classes.push(classes[i]);
        }
    }
    section_image.setAttribute("class", new_classes.join(' '));
}


function _updateSelectionInformation(view) {
    var selectionLabel = "";
    if (view.scintilla) {
        /**
         * @type view {Components.interfaces.ISciMoz}
         */
        var scimoz = view.scintilla.scimoz;
        var selectionStart = scimoz.anchor;
        var selectionEnd = scimoz.currentPos;
        var selectionMode = scimoz.selectionMode;

        if (selectionEnd < selectionStart) {
            // Swap the values around
            var tmp = selectionStart;
            selectionStart = selectionEnd;
            selectionEnd = tmp;
        }

        var count = 0;
        var selection;
        var lineStart, lineEnd;
        if (selectionMode == scimoz.SC_SEL_RECTANGLE) {
            // Block selection mode uses different settings from scimoz API
            selection = [];
            lineStart = scimoz.lineFromPosition(selectionStart);
            lineEnd = scimoz.lineFromPosition(selectionEnd);
            for (var i=lineStart; i <= lineEnd; i++) {
                selectionStart = scimoz.getLineSelStartPosition(i);
                selectionEnd = scimoz.getLineSelEndPosition(i);
                count += (selectionEnd - selectionStart);
                selection.push(scimoz.getTextRange(selectionStart, selectionEnd));
            }
            selection = selection.join("");
            if (count) {
                // Just ensure the selection start and end are different,
                // otherwise we won't update the selection label below.
                // Note: it does not matter what the values are from here on.
                selectionEnd = selectionStart + 1;
            }
        } else {
            lineStart = scimoz.lineFromPosition(selectionStart);
            //Components.interfaces.ISciMoz
            if (selectionMode == scimoz.SC_SEL_LINES) {
                // Line selection mode uses different settings from scimoz API
                lineEnd = scimoz.lineFromPosition(selectionEnd);
                selectionStart = scimoz.getLineSelStartPosition(lineStart);
                selectionEnd = scimoz.getLineSelEndPosition(lineEnd);
            } else {
                // With stream selection, need to check from the position
                // before the end character in order to get the line numbering
                // correct when the cursor is at the start of a line (but the
                // selection does not include this line!)
                lineEnd = scimoz.lineFromPosition(scimoz.positionBefore(selectionEnd));
            }
            count = selectionEnd - selectionStart;
            selection = scimoz.getTextRange(selectionStart, selectionEnd);
        }

        if (selectionStart != selectionEnd) {
            // character count
            selectionLabel = "Sel: " + selection.length + " ch";
            if (selection.length != count) {
                // byte count
                selectionLabel += ", " + count + " by";
            }
            // line count
            selectionLabel += ", " + (Math.abs(lineEnd - lineStart) + 1) + " ln";
        }
    }
    document.getElementById("statusbar-selection").label = selectionLabel;
}

function _updateLineCol(view) {
    if (typeof(view)=='undefined' || !view)
        return;

    try {
        var lineColText = "Ln: " + view.currentLine + " Col: " + view.currentColumn;
        var lineColWidget = document.getElementById('statusbar-line-col');
        lineColWidget.setAttribute('label', lineColText);
    } catch(ex) {
        // not a view that supports these
        _clearLineCol();
        return;
    }

    _updateSelectionInformation(view);

    // Add the lint message updating in a timeout
    if (_updateLintMessageTimer) {
        // Clear the old timer and then we'll start a new one
        clearTimeout(_updateLintMessageTimer);
    }

    // Don't bother updating lint messages if there is no lintBuffer element
    if (typeof(view.lintBuffer)=='undefined' || !view.lintBuffer)
        return;

    _updateLintMessageTimer = setTimeout(_updateLintMessage, 500, view);
}


function _clearLineCol() {
    var lineColWidget = document.getElementById('statusbar-line-col');
    lineColWidget.removeAttribute("label");
    _addMessage(null, "check", 0, false);
}


function _clearCheck() {
    var checkWidget = document.getElementById('statusbar-check');
    checkWidget.setAttribute('image',
        "chrome://komodo/skin/images/icon_check_ok.png");
}

function _updateCheck(view) {
try {
    if (typeof(view)=='undefined' || !view || !view.prefs)
        return;

    // Update the status bar for the current check status.
    var checkWidget = document.getElementById('statusbar-check');

    // Only have linting for some view types (currently only 'editor').
    if (typeof(view.lintBuffer) == "undefined" || !view.lintBuffer)
    {
        checkWidget.setAttribute('collapsed', 'true');
        checkWidget.removeAttribute("image");
        checkWidget.setAttribute("tooltiptext", "Syntax Checking Status");
        return;
    }
    checkWidget.removeAttribute('collapsed');

    // Is linting enabled?
    var checkingEnabled = view.prefs.getBooleanPref("editUseLinting");

    if (typeof(view.lintBuffer)=='undefined' || !view.lintBuffer)
        return;

    // Is there an error in the linter?
    var lintError = view.lintBuffer.errorString;
    if (lintError) {
        checkWidget.setAttribute('image',
            "chrome://komodo/skin/images/icon_check_error.png");
        checkWidget.setAttribute("tooltiptext",lintError);
        return;
    }

    // If so, get the lintResults.
    //   lintResults == null        -> inprogress
    //   lintResults.length == 0    -> ok
    //   lintResults.length > 0     -> errors/warnings
    var lintResults = view.lintBuffer.lintResults;
    if (!lintResults) {
        if (checkingEnabled) {
            checkWidget.setAttribute("tooltiptext",
                                     "Syntax Checking Status: in progress");
            checkWidget.setAttribute('image',
                "chrome://komodo/skin/images/icon_check_inprogress.png");
        } else {
            checkWidget.setAttribute("tooltiptext",
                                     "Automatic syntax checking disabled: shift-click to start");
            checkWidget.setAttribute('image',
                "chrome://komodo/skin/images/icon_check_ok.png");
        }
    } else {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResults(resultsObj, numResultsObj);
        if (numResultsObj.value == 0) {
            checkWidget.setAttribute('image',
                "chrome://komodo/skin/images/icon_check_ok.png");
            checkWidget.setAttribute("tooltiptext",
                                     "Syntax Checking Status: ok");
        } else {
            checkWidget.setAttribute('image',
                "chrome://komodo/skin/images/icon_check_error.png");
            checkWidget.setAttribute("tooltiptext",
                "Syntax Checking Status: " +
                lintResults.getNumErrors() + " error(s), "+
                lintResults.getNumWarnings() + " warning(s)");
        }
    }
} catch(ex) {
    _log.exception(ex);
}
}

function _addMessageObject(sm)
{
    _messageStack.Push(sm);
    //dump("StatusBar: add message: current stack:\n");
    //_messageStack.Dump();

    // Get the latest message and show it in the UI.
    if (_addMessageTimer) {
        clearTimeout(_addMessageTimer);
    }
    _addMessageTimer = setTimeout(_updateMessage, 200);
    // If this new message timesout then schedule an update for then as well.
    if (sm.timeout > 0) {
        // Allow for some inaccuracy in timeout scheduling to ensure that
        // the message has actually timed-out before updating.
        var epsilon = 300;
        setTimeout(_updateMessage, sm.timeout+epsilon);
    }
}

function _updateMessage()
{
    _addMessageTimer = null;
    //dump("StatusBar: update message: current stack:\n");
    //_messageStack.Dump();

    // Get the latest message and show it in the UI.
    var sm = _messageStack.Top();
    var messageWidget = document.getElementById('statusbar-message');
    if (sm) {
        _log.debug("StatusBar: update message: top: msg='"+sm.msg+"', category='"+
             sm.category+"', timeout='"+sm.timeout+"', highlight='"+
             sm.highlight+"'");

        messageWidget.setAttribute("category", sm.category);
        messageWidget.setAttribute("label", sm.msg);
        messageWidget.setAttribute("tooltiptext", sm.msg);
        if (sm.highlight) {
            messageWidget.setAttribute("highlite","true");
        } else {
            messageWidget.removeAttribute("highlite");
        }
    } else {
        messageWidget.setAttribute("label","Ready");
        messageWidget.removeAttribute("tooltiptext");
        messageWidget.removeAttribute("highlite");
    }
}


function _clear() {
    _clearEncoding();
    _clearLanguage();
    _clearLineCol();
    _clearCheck();
}

//---- local classes

function StatusBarObserver() {
    var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
    obsSvc.addObserver(this, 'status_message',false);

    obsSvc.addObserver(this, 'scc_status_message',false);

    obsSvc.addObserver(this, 'autocomplete-did-enter-text',false);
    obsSvc.addObserver(this, 'codeintel_buffer_scanned',false);
    window.addEventListener('current_view_changed',
                            this.handle_current_view_changed, false);
    window.addEventListener('current_view_check_status',
                            this.handle_current_view_check_status, false);
    window.addEventListener('current_view_encoding_changed',
                            this.handle_current_view_encoding_changed, false);
    window.addEventListener('current_view_language_changed',
                            this.handle_current_view_language_changed, false);
    window.addEventListener('current_view_linecol_changed',
                            this.handle_current_view_linecol_changed, false);
    window.addEventListener('view_closed',
                            this.handle_current_view_closed, false);
    ko.main.addWillCloseHandler(this.destroy, this);
};

StatusBarObserver.prototype.destroy = function()
{
    if (_updateLintMessageTimer) {
        clearTimeout(_updateLintMessageTimer);
        _updateLintMessageTimer = null;
    }
    var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
    obsSvc.removeObserver(this, 'status_message');

    obsSvc.removeObserver(this, 'scc_status_message');

    obsSvc.removeObserver(this, 'autocomplete-did-enter-text');
    obsSvc.removeObserver(this, 'codeintel_buffer_scanned');

    window.removeEventListener('current_view_changed',
                               this.handle_current_view_changed, false);
    window.removeEventListener('current_view_check_status',
                               this.handle_current_view_check_status, false);
    window.removeEventListener('current_view_encoding_changed',
                               this.handle_current_view_encoding_changed, false);
    window.removeEventListener('current_view_language_changed',
                               this.handle_current_view_language_changed, false);
    window.removeEventListener('current_view_linecol_changed',
                               this.handle_current_view_linecol_changed, false);
    window.removeEventListener('view_closed',
                               this.handle_current_view_closed, false);
    _messageStack = null;
    _observer = null;
    _prefObserver = null;
}
StatusBarObserver.prototype.observe = function(subject, topic, data)
{
    // Unless otherwise specified the 'subject' is the view, and 'data'
    // arguments are expected to be empty for all notifications.
    _log.debug("StatusBar observed '"+topic+"': ");
    var view = subject;

    switch (topic) {
    case 'autocomplete-did-enter-text':
        // Dev note: We can't do this with an event because this notification is sent from
        // http://mxr.mozilla.org/mozilla-central/source/toolkit/components/autocomplete/src/nsAutoCompleteController.cpp#1158
        
        if (subject != document.getElementById('statusbar-section-textbox')) {
            return;
        }
        
        // Used by the sections list. Only observing when the menulist
        // popup is up.
        //
        // This is a total HACK used to get relevant info about which
        // section in the popup list was selected. The textbox
        // autocomplete controller closes the popup before sending the
        // 'ontextenter' notification that we use to handle a section
        // selection. We need to know the popup's selected index to
        // get the selected index (the row text isn't sufficient to
        // uniquely identify the item).
        ko.statusBar.sectionPopupSelection(subject.popup.selectedIndex);
        break;

    case 'scc_status_message':
        ko.scc.logger.append(data);
        // fall through

    case 'status_message':
        // "subject" is expected to be a koIStatusMessage object.
        _addMessageObject(subject);
        break;
    case 'codeintel_buffer_scanned':
        var buf = subject;
        var currentView = ko.views.manager.currentView;
        var currentPath = null;
        try {
            currentPath = currentView.document.displayPath;
        } catch(ex) {}
        if (currentPath && currentPath == buf.path) {
            _updateSectionPanel(currentView, true);
        }
    }
}

StatusBarObserver.prototype.handle_current_view_changed = function(event) {
    if (!ko.views.manager.batchMode) {
        var view = event.originalTarget;
        _updateEncoding(view);
        _updateLanguage(view);
        _updateLineCol(view);
        _updateSectionPanel(view);
        _updateCheck(view);
    }
};

StatusBarObserver.prototype.handle_current_view_check_status = function(event) {
    _updateCheck(ko.views.manager.currentView);
};

StatusBarObserver.prototype.handle_current_view_encoding_changed = function(event) {
    _updateEncoding(ko.views.manager.currentView);
};

StatusBarObserver.prototype.handle_current_view_language_changed = function(event) {
    _updateLanguage(ko.views.manager.currentView);
    _updateSectionPanel(ko.views.manager.currentView);
};

StatusBarObserver.prototype.handle_current_view_linecol_changed = function(event) {
    _updateLineCol(ko.views.manager.currentView);
    _updateSectionPanel(ko.views.manager.currentView);
};

StatusBarObserver.prototype.handle_current_view_closed = function(event) {
    _clear()
};

function _addMessage(msg, category, timeout, highlight,
                              interactive /* false */)
{
    // Post a message to the status bar message area.
    // "msg" is the message string. An empty string or null indicates
    //      that the message (of the given category) should be cleared.
    // "category" is the message group to which the message belongs. It
    //      is an arbitrary string (it must be at least one character).
    // "timeout" is the amount of time, in milliseconds, that the message
    //      should appear. A value of 0 indicates that the message does
    //      not timeout.
    // "highlight" is a boolean indicating whether the message should be
    //      highlighted on the status bar.
    // "interactive" is a boolean indicating whether the message corresponds
    //      to an interactive prompt (such as interactive search).  These
    //      have higher 'priority' over non-interactive messages in case of
    //      conflict.
    //
    // A structure similar to a stack of status messages is maintained.
    // The latest message is always shown. When/if it timesout then the
    // previous message is the stack is displayed. There can only be one
    // message per category, so reusing a category allows for removal of
    // status messages that are no longer appropriate.
    //
    // To add a message that does not timeout:
    //  _addMessage("hello there", "my_category", 0, false)
    // To remove that message:
    //  _addMessage(null, "my_category", 0, false)
    // To add a highlighted message for three seconds:
    //  _addMessage("hello there", "my_category", 3000, true)
    _log.debug("StatusBar: add message: msg='"+msg+"', category='"+category+
        "', timeout='"+timeout+"', highlight='"+highlight+"'");

    // create a status message component and insert it into the stack
    var sm = Components.classes["@activestate.com/koStatusMessage;1"].
          createInstance(Components.interfaces.koIStatusMessage);
    sm.msg = msg;
    sm.category = category;
    sm.timeout = timeout;
    sm.highlight = highlight;
    if (typeof(interactive) == 'undefined') {
        interactive = false;
    }
    sm.interactive = interactive;
    _addMessageObject(sm);
}

function StatusBarPrefObserver()
{
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    prefSvc.prefs.prefObserverService.addObserver(this,'editUseLinting',0);
    ko.main.addWillCloseHandler(this.destroy, this);
};

StatusBarPrefObserver.prototype.destroy = function()
{
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    prefSvc.prefs.prefObserverService.removeObserver(this,'editUseLinting');
}
StatusBarPrefObserver.prototype.observe = function(prefSet, prefName, prefSetID)
{
    _log.debug("StatusBar: observed pref '"
            + prefName + "' change (prefSet="
            + prefSet + ", prefSetID="
            + prefSetID)
    if (prefName == "editUseLinting") {
        var view = ko.views.manager.currentView;
        if (view) {
            _updateCheck(view);
        }
    }
};

_observer = new StatusBarObserver();
_prefObserver = new StatusBarPrefObserver();

//---- public functions

this.sectionMousedown = function(event) {
    if (ko.views.manager.currentView) {
        var section_panel = document.getElementById("statusbar-section");
        if (section_panel.getAttribute('disabled')) return;
        this.sectionActivate();
    }
}

this.sectionBlur = function() {
    if (this._sectionActive) {
        this.sectionDeactivate();
    }
}

this.sectionReverted = function() {
    this.sectionDeactivate();
}

this.sectionActivate = function() {
    try {
        var panel = document.getElementById("statusbar-section");
        var deck = document.getElementById("statusbar-section-deck");
        var textbox = document.getElementById("statusbar-section-textbox");
        
        deck.selectedIndex = 1;
        this._sectionSelectedIndex = -1;
        textbox.value = "";
        textbox.focus();
        // Open the complete list by faking a <down arrow> keypress
        // in the autocomplete textbox.
        textbox.controller.handleKeyNavigation(Components.interfaces.nsIDOMKeyEvent.DOM_VK_DOWN);

        this._sectionActive = true;
        this._sectionHaveDoneFirstSearch = false;
    } catch(e) {
        _log.exception(e);
    }
}

this.sectionDeactivate = function() {
    try {
        var deck = document.getElementById("statusbar-section-deck");
        deck.selectedIndex = 0;
        ko.views.manager.currentView.setFocus();
        this._sectionActive = false;
    } catch(e) {
        _log.exception(e);
    }
}

this.sectionSearchComplete = function() {
    try {
        if (!this._sectionHaveDoneFirstSearch) {
            // For the first search (without a filter) ensure that the
            // current section is visible in the popup tree. The *best*
            // way to handle this would be to get the current index from
            // the last `KoTACSectionsSearch` search. However, the
            // textbox autocomplete controller doesn't provide a way to
            // get the return search result where we could store other
            // relevant information.
            //
            // As a workaround we'll hack it by getting the same
            // information from the buffer. This presumes that (a) the
            // search *was* with an empty searchString and (b) that the
            // current line number hasn't changed.
            this._sectionHaveDoneFirstSearch = true;

            var view = ko.views.manager.currentView;
            var ciBuf = view.document.ciBuf;
            var currIdx = ciBuf.curr_section_idx_from_line(view.currentLine);

            var textbox = document.getElementById("statusbar-section-textbox");
            var tree = textbox.popup.tree.treeBoxObject; // nsITreeBoxObject
            
            var matchCount = textbox.controller.matchCount;
            var pageLen = tree.getPageLength();
            if (pageLen < matchCount) {
                var topIdx = Math.floor(currIdx - (pageLen/2));
                // Make sure don't scroll so far that there is whitespace
                // in the popup below the last section.
                topIdx = Math.min(topIdx, matchCount-pageLen);
                topIdx = Math.max(0, topIdx);
                tree.scrollToRow(topIdx);
            }
        }
    } catch(e) {
        _log.exception(e);
    }
}

this.sectionPopupSelection = function(selectedIndex) {
    this._sectionSelectedIndex = selectedIndex;
}

/* 'ontextentered' on the autocomplete-textbox indicates that a selection
 * has been made and that we want to jump to the corresponding line.
 */
this.sectionTextEntered = function() {
    var deck = document.getElementById("statusbar-section-deck");
    var textbox = document.getElementById("statusbar-section-textbox");

    // Determine the line to jump to, if any.
    var line = -1;  // -1 means no line to jump to
    if (this._sectionSelectedIndex == -1) {
        // This means a selection was NOT made from the popup list. I.e.
        // <Enter> was hit in the textbox with no selection in the popup
        // list. Use the first match.
        if (textbox.controller.matchCount) {
            line = Number(textbox.controller.getCommentAt(0));
        } else {
            ko.statusBar.AddMessage("no sections match '"+textbox.value+"'",
                                    "sections", 3000, true);
        }
    } else {
        line = Number(textbox.controller.getCommentAt(this._sectionSelectedIndex));
    }
    
    // Do the jump.
    if (line != -1) {
        var view = ko.views.manager.currentView;
        view.scimoz.gotoLine(line-1);  // 1-based -> 0-based
        view.setFocus();
    }
    
    this.sectionDeactivate();
}



/**
 * dump
 *
 * dump the current message stack to stdout
 */
this.dump = function() { _messageStack.Dump(); }

/**
 * ClearCheck
 *
 * clear the syntax checking status
 */
this.ClearCheck = function() { _clearCheck(); }

/**
 * AddMessage
 *
 * Post a message to the status bar message area.
 * "msg" is the message string. An empty string or null indicates
 *      that the message (of the given category) should be cleared.
 * "category" is the message group to which the message belongs. It
 *      is an arbitrary string (it must be at least one character).
 * "timeout" is the amount of time, in milliseconds, that the message
 *      should appear. A value of 0 indicates that the message does
 *      not timeout.
 * "highlight" is a boolean indicating whether the message should be
 *      highlighted on the status bar.
 * "interactive" is a boolean indicating whether the message corresponds
 *      to an interactive prompt (such as interactive search).  These
 *      have higher 'priority' over non-interactive messages in case of
 *      conflict.
 *
 * A structure similar to a stack of status messages is maintained.
 * The latest message is always shown. When/if it timesout then the
 * previous message is the stack is displayed. There can only be one
 * message per category, so reusing a category allows for removal of
 * status messages that are no longer appropriate.
 *
 * To add a message that does not timeout:
 *  ko.statusBar.addMessage("hello there", "my_category", 0, false)
 * To remove that message:
 *  ko.statusBar.addMessage(null, "my_category", 0, false)
 * To add a highlighted message for three seconds:
 *  ko.statusBar.addMessage("hello there", "my_category", 3000, true)
 */
this.AddMessage = function(msg, category, timeout, highlight, interactive)
    { _addMessage(msg, category, timeout, highlight, interactive); }

/**
 * Clear
 *
 * clear all statusbar elements
 */
this.Clear = function() { _clear(); }

}).apply(ko.statusBar);


// backwards compatible APIs
var StatusBar = ko.statusBar;
var StatusBar_AddMessage = ko.statusBar.AddMessage;


